#include "..\Precomp.h"
#include "World.h"
#include "GameObj.h"
#include "Player.h"
#include "..\Network\Network.h"

CWorld g_World;

// see Defines.h for color enum
static byte ResColors[RC_NUM*3] =
{ 0,0,0,      96,96,96,   128,128,128,  255,255,255,
  128,0,0,    192,0,0,    255,0,0,      0,96,0,
  0,128,0,    0,255,0,    0,0,128,      0,0,192,
  0,0,255,    128,0,128,  255,0,255,    0,128,128,
  0,192,192,  0,255,255,  255,128,0,    128,128,0,
  255,255,0,  64,32,16,   128,64,32
};

// world states
enum
{ WS_MULTI, WS_SINGLE, WS_BASEMENU, WS_DEAD, WS_MAYBEQUIT, WS_WAITSPAWN,
  WS_TRYCONNECT, WS_CONNLOST, WS_CONNFAIL, WS_LOGINFAIL
};

CWorld::CWorld()
{ Map      = NULL;
  Players  = NULL;
  Objects  = NULL;
  Circs    = NULL;
  Camera   = NULL;
  pDPC     = NULL;
  hMoveMsg = NULL;
  Gravity  = 0.01666666666667;
  nObjects = nPlayers   = nCircs = nLines     = 0;
  maxCircs = maxObjects = nLines = maxPlayers = 0;
  bLocked  = bQuit = false;

  LARGE_INTEGER LI;
  QueryPerformanceFrequency(&LI);
  tInc  = LI.QuadPart / TICKS_PER_SEC;

  InitializeCriticalSection(&CS);
}

CWorld::~CWorld()
{ Unload();
  if(pDPC)
  { pDPC->Release();
    pDPC=NULL;
  }
  delete[] Objects;
  delete[] Players;
  for(int i=0; i<maxLines; i++) Lines[i].~Line_s();
  operator delete(Lines);
  DeleteCriticalSection(&CS);
}

int CWorld::InitNet()
{ if(pDPC) return 0;

  HRESULT hRet;
  IDirectPlay8Client *dpc;

  CHKRET(DirectPlay8Create(&IID_IDirectPlay8Client, (void**)&dpc, NULL), Done);
  IFS(dpc->Initialize(this, Handler, 0)) pDPC=dpc;
  else dpc->Release();

  Done:
  return FAILED(hRet);
}

int CWorld::Load(const char *fore, const char *back)
{ Unload();

  State = WS_SINGLE;
  bNet  = false;
  return LoadLevels(fore, back);
}

int CWorld::LoadNet(const char *ip, const char *name)
{ Unload();

  Name = name;

  State = WS_TRYCONNECT, bNet = true;
  if(InitNet()) return 1;
  g_Font.PrintF(g_Back, 60, 148, RC_WHITE, "Connecting to %s...", ip);
  g_Back->Flip();
  if(Connect(ip)) return 2;
  return 0;
}

void CWorld::Unload()
{ int i;
  if(bNet) Disconnect();
  for(i=0; i<nPlayers; i++) Players[i]->Destroy();
  for(i=0; i<nObjects; i++) Objects[i]->Destroy();

  delete Back;
  delete Map;
  Map = Back = NULL;
  nPlayers   = nObjects = 0;
}

void CWorld::SetTiming()
{ LARGE_INTEGER LI;
  QueryPerformanceCounter(&LI);
  tNext = LI.QuadPart + tInc;
}

void CWorld::DoTicks()
{ LARGE_INTEGER LI;
  QueryPerformanceCounter(&LI);
  tCur = LI.QuadPart;
  while(tCur >= tNext && !bQuit)
  { DoInput();
    Advance();
    tNext += tInc;
  }
}

void CWorld::Advance()
{ switch(State)
  { case WS_MULTI:     case WS_SINGLE: case WS_BASEMENU:
    case WS_MAYBEQUIT: case WS_DEAD:   case WS_WAITSPAWN:
      AdvanceWorld();
      break;
  }
}


void CWorld::Render()
{ switch(State)
  { case WS_MULTI: case WS_SINGLE: case WS_WAITSPAWN:
      RenderWorld();
      break;

    case WS_BASEMENU:
      RenderWorld();
      g_Font.PrintF(g_Back, 10, 120, bMenuGun ? RC_WHITE : RC_LTGREY,
                    "Gun:     %s", g_WepName[nGun]);
      g_Font.PrintF(g_Back, 10, 120+g_Font.FontHei+2, bMenuGun ? RC_LTGREY : RC_WHITE,
                    "Special: %s", g_WepName[nSpecial]);
      break;

    case WS_DEAD:
      RenderWorld();
      g_Font.Print(g_Back, "Press [space] to respawn!", 60, 148, RC_WHITE);
      break;

    case WS_MAYBEQUIT:
      RenderWorld();
      g_Font.Print(g_Back, "Really quit? (Enter=yes, Esc=no)", 60, 148, RC_WHITE);
      break;
  }
}

int CWorld::Connect(const char *ip)
{ assert(pDPC);
  DPN_APPLICATION_DESC   ad;
  IDirectPlay8Address   *addr, *dev;
  HRESULT                hRet;
  DPNHANDLE              dummy;
  WCHAR                  adbuf[512];

  memset(&ad, 0, sizeof(ad));
  ad.dwSize = sizeof(ad);
  ad.guidApplication = g_AppGuid;

  DirectPlay8AddressCreate(&IID_IDirectPlay8Address, (void**)&addr, NULL);
  wcscpy(adbuf, L"x-directplay:/provider=%7BEBFE7BA0-628D-11D2-AE0F-006097B01411%7D;"
                L"port=" NETPORTSTR L";hostname=");
  cw_cat(adbuf, ip);
  addr->BuildFromURLW(adbuf);

  DirectPlay8AddressCreate(&IID_IDirectPlay8Address, (void**)&dev, NULL);
  dev->BuildFromURLW(L"x-directplay:/provider=%7BEBFE7BA0-628D-11D2-AE0F-006097B01411%7D");

  hRet = pDPC->Connect(&ad, addr, dev, NULL, NULL, NULL, 0, NULL, &dummy, 0);

  addr->Release();
  dev->Release();
  return FAILED(hRet);
}

void CWorld::Disconnect()
{ if(pDPC) pDPC->Close(0);
}

int CWorld::Send(NetMsg *msg, DWORD flags, DWORD msTimeout, DPNHANDLE *hand)
{ assert(pDPC);
  DPN_BUFFER_DESC desc;
  DPNHANDLE       dummy;
  flags |= ((flags&NET_SENDGUAR) || hand) ? 0 : DPNSEND_NOCOMPLETE;

  desc.dwBufferSize = msg->Len;
  desc.pBufferData  = (byte*)msg;
  return FAILED(pDPC->Send(&desc, 1, msTimeout, NULL, hand ? hand : &dummy, flags));
}

void CWorld::MkExp(ExpType type, int x, int y, ShipObj *s, int seed)
{ if(s==&Me->Ship)
  { g_SRand(seed=rand());
    if(bNet)
    { MsgExplode msg;
      msg.EType = type;
      msg.ID    = s->ID;
      msg.X     = x, msg.Y = y;
      msg.Seed  = seed;
      Send(&msg, NET_SENDGUAR);
    }
  }
  else g_SRand(seed);
  Explosion(type, x, y, s);
}

void CWorld::Explosion(ExpType type, int x, int y, ShipObj *s)
{ enum
  { DOT, S_EXP, M_EXP, L_EXP
  };
  static struct
  { int Lows [4];
    int Highs[4];
    int Rad;
  } Stats[ET_NUM] =
  { { { 0,  1,  0,  0 },  { 2,   1,  1,  0  }, 4  },
    { { 3,  1,  1,  0 },  { 6,   3,  3,  0  }, 10 },
    { { 10, 4,  2,  1 },  { 18,  6,  5,  2  }, 20 },
    { { 18, 7,  4,  2 },  { 25,  11, 8,  5  }, 14 },
    { { 25, 10, 8,  4 },  { 40,  18, 14, 8  }, 50 },
    { { 50, 20, 16, 8 },  { 150, 36, 28, 16 }, 70 }
  };

  Vector<FLOAT> pos(x, y);
  if(!s) s = &Me->Ship; // FIXME
  int i, j, a=0, num, rad, ang, dist[4];
  AmmoBullet *b;
  ExplodeObj *e;

  for(i=0; i<4; i++) dist[i] = Stats[type].Highs[i]-Stats[type].Lows[i];

  num = Stats[type].Lows[DOT];
  if(dist[DOT]) num += g_Rand()%dist[DOT];
  rad = Stats[type].Rad;
  for(i=0; i<num; i++)
  { b=(AmmoBullet*)g_AmmoFact.Create(AT_BULLET);
    b->Pos = pos+g_Vects[ang=(byte)g_Rand()]*(g_Rand()%rad);
    b->Vel = g_Vects[ang];
    b->Weight = 4, b->Color = s->Color, b->Owner = s->ID;
    b->LoadShip(s);
    AddObject(b);
  }

  for(i=S_EXP; i<=L_EXP; i++)
  { num = Stats[type].Lows[i];
    if(dist[i]) num += g_Rand()%dist[i];
    a+=num;
    for(j=0; j<num; j++)
    { e=ExplodeObj::Alloc.New();
      if(type == ET_TINY) e->Pos = pos;
      else
      { e->Pos = pos+g_Vects[ang=(byte)g_Rand()]*(g_Rand()%8);
        e->Vel = g_Vects[ang]*(g_Rand()%(type+1)*0.5);
        e->Activate = g_Rand()%(a<<2);
      }
      e->Owner = s->ID, e->Type = i-1;
      AddObject(e);
    }
  }
}

void CWorld::AddExp(ExplodeObj *o)
{ int x=(int)o->Pos.X, y=(int)o->Pos.Y, rad=o->Types[o->Type].Radius;
  if(nCircs == maxCircs)
  { CircleRem *tmp = NEW CircleRem[maxCircs = maxCircs ? maxCircs*2 : 16];
    if(nCircs)
    { memcpy(tmp, Circs, nCircs*sizeof(CircleRem));
      delete[] Circs;
    }
    Circs = tmp;
  }
  for(int i=0; i<nCircs; i++)
    if(Circs[i].X == x && Circs[i].Y == y && Circs[i].Rad == rad) return;
  Circs[nCircs].X = x, Circs[nCircs].Y = y, Circs[nCircs].Rad = rad;
  nCircs++;
}

void CWorld::DestroyMap(ExplodeObj *o)
{ if(o->Owner==Me->Ship.ID)
  { if(RemoveCircle((int)o->Pos.X, (int)o->Pos.Y, o->Types[o->Type].Radius) && bNet)
      AddExp(o);
  }
}

void CWorld::AddObject(Object *o)
{ if(nObjects == maxObjects)
  { Object **tmp = NEW Object*[maxObjects+=256];
    if(Objects)
    { memcpy(tmp, Objects, nObjects*sizeof(Object*));
      delete[] Objects;
    }
    Objects = tmp;
  }
  Objects[o->Index=nObjects++]=o;
}

void CWorld::RemObject(Object *o)
{ register int i=o->Index;
  if(i != --nObjects) (Objects[i]=Objects[nObjects])->Index=i;
}

int CWorld::FindObject(unsigned id)
{ assert(id);
  for(int i=0; i<nObjects; i++) if(Objects[i]->ID==id) return i;
  return -1;
}

void CWorld::AddPlayer(Player *p)
{ if(nPlayers == maxPlayers)
  { Player **tmp = NEW Player*[maxPlayers+=16];
    if(Players)
    { memcpy(tmp, Players, nPlayers*sizeof(Player*));
      delete[] Players;
    }
    Players = tmp;
  }
  if(!nPlayers) SetCamera(&p->Ship);
  Players[nPlayers++]=p;
}

void CWorld::RemPlayer(Player *p)
{ assert(&p->Ship != Camera);
  for(int i=0; i<nPlayers; i++)
  { if(Players[i]==p)
    { Players[i] = Players[--nPlayers];
      return;
    }
  }
  assert(0);
}

int CWorld::FindPlayer(unsigned id)
{ for(int i=0; i<nPlayers; i++) if(Players[i]->Ship.ID==id) return i;
  return -1;
}

void CWorld::DoInput()
{ Keypress kp;
  switch(State)
  { case WS_MULTI: case WS_SINGLE:
      if(g_KB.Keys[DIK_LEFT])
      { Me->Ship.Rotate(-Me->Ship.TurnAcc++);
        if(Me->Ship.TurnAcc>3) Me->Ship.TurnAcc=3;
      }
      if(g_KB.Keys[DIK_RIGHT])
      { Me->Ship.Rotate(Me->Ship.TurnAcc++);
        if(Me->Ship.TurnAcc>3) Me->Ship.TurnAcc=3;
      }
      if(g_KB.Keys[DIK_UP])     Me->Ship.Accel();
      else                      Me->Ship.Flag.bAccel=false;
      while(g_KB.GetData(&kp))
      { if(kp.Key==DIK_A)      Me->Ship.CtlFire(GP_GUN,     kp.Down!=0);
        else if(kp.Key==DIK_S)
        { if(kp.Down && Me->Ship.Flag.bOnBase)
          { Me->Ship.CtlFire(GP_SPECIAL, false);
            nGun = Me->Ship.nGun, nSpecial = Me->Ship.nSpecial;
            bMenuGun = true;
            SetState(WS_BASEMENU);
          }
          else Me->Ship.CtlFire(GP_SPECIAL, kp.Down!=0);
        }
        else if(kp.Key==DIK_ESCAPE && !kp.Down) SetState(WS_MAYBEQUIT);
      }
      break;
    
    case WS_BASEMENU:
      while(g_KB.GetData(&kp))
      { if(kp.Down)
        { if(kp.Key==DIK_A || kp.Key==DIK_S)
          { if(nGun != Me->Ship.nGun || nSpecial != Me->Ship.nSpecial)
            { if(bNet)
              { MsgWeaponSel msg;
                msg.Gun = (WepType)nGun, msg.Special = (WepType)nSpecial;
                msg.ID  = Me->Ship.ID;
                Send(&msg, NET_SENDGUAR);
              }
              Me->Ship.SetWeps((WepType)nGun, (WepType)nSpecial);
            }
            SetState(OldState);
          }
        }
        else
        { if(kp.Key==DIK_UP || kp.Key == DIK_DOWN) bMenuGun = !bMenuGun;
          else if(kp.Key==DIK_LEFT)
          { if(bMenuGun)
            { if(--nGun < 0) nGun=WT_GUNMAX;
            }
            else if(--nSpecial <= WT_GUNMAX) nSpecial=WT_SPECMAX;
          }
          else if(kp.Key==DIK_RIGHT)
          { if(bMenuGun)
            { if(nGun++ == WT_GUNMAX) nGun=0;
            }
            else if(nSpecial++ == WT_SPECMAX) nSpecial=WT_GUNMAX+1;
          }
          else if(kp.Key==DIK_ESCAPE) SetState(OldState);
        }
      }
      break;

    case WS_DEAD:
      while(g_KB.GetData(&kp))
      { if(!kp.Down) continue;
        if(kp.Key==DIK_SPACE)
        { SetState(WS_WAITSPAWN);
          MsgGetSpawn msg;
          Send(&msg, NET_SENDGUAR);
        }
        else if(kp.Key==DIK_ESCAPE) SetState(WS_MAYBEQUIT);
      }
      break;

    case WS_WAITSPAWN:
      while(g_KB.GetData(&kp))
        if(kp.Down==false && kp.Key==DIK_ESCAPE) SetState(WS_MAYBEQUIT);
      break;

    case WS_MAYBEQUIT:
      while(g_KB.GetData(&kp))
      { if(kp.Down) continue;
        if(kp.Key==DIK_RETURN)
        { bQuit = true;
          g_Win.Quit();
        }
        else if(kp.Key==DIK_ESCAPE) State = OldState;
      }
      break;
  }
}

void CWorld::AddLine(const char *str, ...)
{ va_list list;
  int  len;
  char buf[1024];

  va_start(list, str);
  len=vsprintf(buf, str, list)*6;
  va_end(list);

  Lock();
  if(nLines==maxLines)
  { Line_s *tmp = (Line_s*)oper_new((maxLines+=16)*sizeof(Line_s));
    if(nLines)
    { memcpy(tmp, Lines, nLines*sizeof(Line_s));
      operator delete(Lines);
    }
    for(int i=nLines; i<maxLines; i++) place_new(tmp+i) Line_s;
    Lines = tmp;
  }
  Lines[nLines].Line   = buf;
  Lines[nLines].Age    = 0;
  Lines[nLines].MaxAge = max(len, TICKS_PER_SEC);
  nLines++;
  Unlock();
}

int CWorld::LoadLevels(const char *fore, const char *back)
{ if(!Objects) Objects = NEW Object*[maxObjects=256];
  if(!Players) Players = NEW Player*[maxPlayers=8];

  if(back)
  { Back = g_VFS.LoadPcx(back, NULL, CDD_SYSTEM);
    if(!back) return 1;
  }
  Map = g_VFS.LoadPcx(fore, NULL, CDD_SYSTEM, Palette);
  if(!Map) return 2;

  memset(Palette, 0, 3);
  memcpy(Palette+(RC_START*3), ResColors, RC_NUM*3);
  if(g_Win.bFocus && !g_Win.bMinim) g_DD.SetPal(Palette);
  g_DD.InitAlpha(Palette);

  Map->Stats.Trans=true;
  Width=Map->Width, Height=Map->Height, Tick=0;

  return 0;
}

void CWorld::AdvanceWorld()
{ Object  *o;
  ShipObj *s;
  int      i, j, len;

  Map->Lock();

  Lock();
  bLocked = true;
  if(nLines && Lines[0].Age++==Lines[0].MaxAge)
  { char buf[sizeof(Line_s)];
    if(--nLines)
    { memcpy(buf, Lines, sizeof(Line_s));
      memcpy(Lines, Lines+nLines, sizeof(Line_s));
      memcpy(Lines+nLines, buf, sizeof(Line_s));
    }
  }

  for(i=0; i<nObjects; i++)
  { if(Objects[i]->Flag.bHidden) continue;
    Objects[i]->Tick(Tick);
    Objects[i]->CalcVel();
    Objects[i]->Move();
  }

  for(i=0; i<nPlayers; i++)
  { if(!Players[i]->Ship.bAlive) continue;
    Players[i]->Ship.Shoot();
    Players[i]->Ship.Tick(Tick);
    Players[i]->Ship.CalcVel();
    Players[i]->Ship.Move();
  }

  for(i=0; i<nObjects; i++)
  { o = Objects[i];
    o->Age++;
    if(o->Flag.bHidden) continue;
    if(o->IsHitMap()) o->HitMap();
    if(o->Flag.bHitObjs)
    { for(j=0; j<nObjects; j++)
      { if(i==j || Objects[j]->Flag.bHidden) continue;
        if(o->Intersects(Objects[j])) o->HitObject(Objects[j]);
      }
    }
    if(o->Flag.bDead) continue;
    for(j=0; j<nPlayers; j++)
      if(Players[j]->Ship.bAlive && o->Intersects(&Players[j]->Ship))
        o->HitShip(&Players[j]->Ship);
  }

  for(i=0; i<nObjects; i++)
  { if((o=Objects[i])->Flag.bDead)
    { if(o->ID && o->Owner!=Me->Ship.ID) o->Flag.bHidden=true;
      else
      { RemObject(o);
        o->Destroy();
        i--;
      }
    }
  }

  if(!(Tick & 0x3FF))
  { for(i=0; i<nObjects; i++)
    { if((o=Objects[i])->Flag.bDead && o->Age>=OBJ_KILL_AGE)
      { RemObject(o);
        o->Destroy();
        i--;
      }
    }
  }

  for(i=0; i<nPlayers; i++)
  { s = &Players[i]->Ship;
    if(s->IsHitMap()) s->HitMap();
    else
    { s->Flag.bOnBase = s->Flag.bRest = false;
      if(Me==Players[i]) s->OldPos = s->Pos;
    }
    /*for(j=0; j<nPlayers; j++)
    { if(i==j) continue;
      if(s->Intersects(&Players[j]->Ship)) s->HitShip(&Players[j]->Ship);
    }*/
  }

  if(Me->Ship.Health<=0 && Me->Ship.bAlive)
  { SetState(WS_DEAD);
    MsgDeath msg;
    msg.ID = Me->Ship.ID;
    Send(&msg, NET_SENDGUAR);

    if(Me->Ship.Health <= -50)
      MkExp(ET_MEDIUM, (int)Me->Ship.Pos.X, (int)Me->Ship.Pos.Y, &Me->Ship);
    else
    { char buf[sizeof(MsgNewObject)+4];
      MsgNewObject *msg = place_new(buf) MsgNewObject;
      DeadShipObj::DStruct *dat = (DeadShipObj::DStruct*)&msg->ObjData;
      msg->OType  = OT_DEADSHIP;
      msg->ID     = Me->Ship.ID;
      msg->X      = (int)Me->Ship.Pos.X, msg->Y = (int)Me->Ship.Pos.Y;
      dat->Health = Me->Ship.Health;
      msg->Len   += sizeof(DeadShipObj::DStruct);
      Send(msg, NET_SENDGUAR);
    }
    Me->Ship.bAlive=false;
  }

  if(bNet)
  { for(i=0; i<nPlayers; i++)
    { if(Players[i] != Me && Players[i]->Ship.Flag.bAccel)
        Players[i]->Ship.AddFlame();
    }
    if(Me->Ship.bAlive) SendMove();
    if(!(Tick&3))
    { byte buf[1024];
      i=0;
      while(i<nCircs)
      { MsgRemCircles *msg = (MsgRemCircles *)place_new(buf) MsgRemCircles;
        j = min(nCircs-i, (sizeof(buf)-sizeof(MsgRemCircles))/sizeof(CircleRem));
        msg->NumCircs = j;
        memcpy(msg->Circs, Circs+i, len=j*sizeof(CircleRem));
        msg->Len += len;
        Send(msg, NET_SENDGUAR);
        i += j;
      }
      nCircs = 0;
    }
  }

  bLocked = false;
  Tick++;
  Unlock();
  Map->Unlock();
}

void CWorld::RenderWorld()
{ Lock();

  int     w = g_Win.Width/2,      h = g_Win.Height/2, i;
  int     x = (int)Camera->Pos.X, y = (int)Camera->Pos.Y;
  RECT rect = { x-w, y-h, x+w, y+h-1 }, back;

  if(rect.left < 0) rect.right -= rect.left, rect.left=0;
  else if(rect.right > Map->Width)
    rect.left -= rect.right-Map->Width, rect.right=Map->Width;

  if(rect.top < 1) rect.bottom -= rect.top-1, rect.top=1;
  else if(rect.bottom > Map->Height)
    rect.top -= rect.bottom-Map->Height, rect.bottom=Map->Height;

  if(Back)
  { back.top    = back.left = 0;
    back.right  = g_Win.Width;
    back.bottom = 1;
    g_Back->Fill(&back);

    if((w=Map ->Width-g_Win.Width)<0) w=0;
    if((h=Back->Width-g_Win.Width)<0) h=0;
    back.left   = w&&h ? rect.left*h/w : 0;
    back.right  = back.left+g_Win.Width;
    
    if((w=Map ->Height-g_Win.Height+1)<0) w=0;
    if((h=Back->Height-g_Win.Height+1)<0) h=0;
    back.top    = w&&h ? rect.top*h/w : 0;
    back.bottom = back.top+g_Win.Height-1;

    //if(back.right  > Back->Width)  back.right  = Back->Width;
    //if(back.bottom > Back->Height) back.bottom = Back->Height;

    Back->BlitFast(g_Back, 0, 1, &back);
  }
  else g_Back->Fill();

  Map->Lock();
  g_Back->Lock();

  for(i=0; i<nObjects; i++)
    if(!Objects[i]->Flag.bHidden &&
       PtInRect(&rect, (int)Objects[i]->Pos.X, (int)Objects[i]->Pos.Y))
      Objects[i]->Draw(rect.left, rect.top);

  for(i=0; i<nPlayers; i++)
    if(Players[i]->Ship.bAlive &&
       PtInRect(&rect, (int)Players[i]->Ship.Pos.X, (int)Players[i]->Ship.Pos.Y))
        Players[i]->Draw(rect.left, rect.top);

  w = g_Win.Width>>2;
  if(Me->Ship.Health)
    Line2D(g_Back, Vertex2DC(0, 0, RC_RED), 
           Vertex2DC(w*Me->Ship.Health/Me->Ship.MaxHealth, 0, RC_LTRED));
  if(Me->Ship.Fuel)
    Line2D(g_Back, Vertex2DC(w, 0, RC_DKGREEN), 
           Vertex2DC(w*Me->Ship.Fuel/Me->Ship.MaxFuel+w, 0, RC_LTGREEN));
  h=w+w;
  if(Me->Ship.Gun && Me->Ship.Gun->Ammo)
    Line2D(g_Back, Vertex2DC(h, 0, RC_DKCYAN), 
           Vertex2DC(w*Me->Ship.Gun->Ammo/Me->Ship.Gun->MaxAmmo+h, 0,
           RC_CYAN));
  h+=w;
  if(Me->Ship.Special && Me->Ship.Special->Ammo)
    Line2D(g_Back, Vertex2DC(h, 0, RC_DKGREY), 
           Vertex2DC(w*Me->Ship.Special->Ammo/Me->Ship.Special->MaxAmmo+h, 0,
           RC_WHITE));

  g_Back->Unlock();
  Map->Unlock();
  Map->BlitFast(g_Back, 0, 1, &rect);

  if(Me->Ship.Flag.bOnBase)
  { g_Font.Print(g_Back, "Press [special] to activate menu!",
                 10, g_Win.Height-g_Font.FontHei*2-5, RC_WHITE);
  }
  if(nLines)
  { g_Font.Print(g_Back, Lines[0].Line.c_str(),
                 10, g_Win.Height-g_Font.FontHei-3, RC_WHITE);
  }

  if(pDPC)
  { DWORD num, bytes;
    pDPC->GetSendQueueInfo(&num, &bytes, 0);
    g_Font.PrintF(g_Back, 3, 2, RC_CYAN, "%d msgs, %d bytes", num, bytes);
  }

  Unlock();
}

void CWorld::SendMove()
{ MsgMove msg;
  msg.X  = Me->Ship.Pos.X, msg.Y  = Me->Ship.Pos.Y;
  msg.XV = Me->Ship.Vel.X, msg.YV = Me->Ship.Vel.Y;
  msg.Angle = (byte)Me->Ship.Angle, msg.Accel = Me->Ship.Flag.bAccel;
  if(hMoveMsg) pDPC->CancelAsyncOperation(hMoveMsg, DPNCANCEL_SEND);
  Send(&msg, 0, 100, &hMoveMsg);
  //Send(&msg, NET_HIGHPRIO, 100);
}

bool CWorld::RemoveCircle(int x, int y, int rad)
{ if(y+rad < 0 || y-rad>=Map->Height ||
     x+rad < 0 || x-rad>=Map->Width)
      return false;

  bool locked = Map->Stats.Locked, ret;
  int dx=0, dy=rad, s=3-2*rad, pitch, wid=Map->Width;

  if (!locked && Map->Lock() != CDD_NOERROR) return true;
  pitch = Map->Pitch;

  Region.Resize(rad*2+1);
  Region.SetRange(INT_MIN, INT_MAX);

  for(;dx<=dy;dx++)
  { Region[rad-dx].SetMin(x-dy);
    Region[rad-dx].SetMax(x+dy);
    if(dx)
    { Region[rad+dx].SetMin(x-dy);
      Region[rad+dx].SetMax(x+dy);
    }
    Region[rad-dy].SetMin(x-dx);
    Region[rad-dy].SetMax(x+dx);
    Region[rad+dy].SetMin(x-dx);
    Region[rad+dy].SetMax(x+dx);
    if(s<0) s+=4*dx+6;
    else
    { s+=4*(dx-dy)+10;
      dy--;
    }
  }

  dx=0, dy=rad*2;
  if(y-rad < 0) dx=rad-y;
  if(y+rad >= Map->Height) dy-=y+rad-Map->Height;
  ret = FlatRegion(y-rad+dx, dx, dy);

  if (!locked) Map->Unlock();
  return ret;
}

bool CWorld::FlatRegion(int y, int start, int end)
{ int pitch=Map->Pitch, wid=Map->Width;
  byte *mem=Map->Mem+pitch*y;
  bool  ret=false;
  for(;start<=end;start++)
  { if(FlatSpan8C(mem, Region[start].min, Region[start].max, wid)) ret=true;
    mem+=pitch;
  }
  return ret;
}

bool CWorld::FlatSpan8C(byte *mem, int start, int end, int wid)
{ byte c;
  bool ret=false;
  if(end<0 || start>=wid) return ret;
  if(start<0) start=0;
  if(end>=wid) end=wid-1;
  mem+=start;
  while(start<=end)
  { c = *mem;
    if(c<224)
    { *mem++=0;
      ret=true;
    }
    else mem++;
    start++;
  }
  return ret;
}

void CWorld::MsgHandler(NetMsg *msg)
{ union
  { NetMsg          *msg;
    MsgLogin        *login;
    MsgLoginSucc    *logsucc;
    MsgLoginFail    *logfail;
    MsgNetConn      *netconn;
    MsgPlayerJoin   *pjoin;
    MsgPlayerLeave  *pleave;
    MsgPlayerMove   *pmove;
    MsgWeaponFire   *wfire;
    MsgWeaponStart  *wstart;
    MsgWeaponEnd    *wend;
    MsgWeaponSel    *wsel;
    MsgObjectHit    *ohit;
    MsgFWeaponHit   *ofhit;
    MsgRemCircles   *rcirc;
    MsgSpawn        *spawn;
    MsgDeath        *death;
    MsgNewObject    *nobj;
    MsgRemObject    *robj;
  } r, s;
  r.msg=msg;

  LoginPlayer *lp;
  Player      *p;
  Object      *o;
  AmmoBase    *ammo;
  int          i, len;
  char        *c, buf[1024];

  s.msg=(NetMsg*)buf;
  c=buf;

  Lock();

  switch(msg->Type)
  { case MT_SCPLAYERMOVE:
      i = FindPlayer(r.pmove->ID);
      Players[i]->Ship.OldPos = Players[i]->Ship.Pos;
      Players[i]->Ship.Pos.X = r.pmove->X;
      Players[i]->Ship.Pos.Y = r.pmove->Y;
      Players[i]->Ship.Vel.X = r.pmove->XV;
      Players[i]->Ship.Vel.Y = r.pmove->YV;
      Players[i]->Ship.Angle = r.pmove->Angle;
      Players[i]->Ship.Flag.bAccel = r.pmove->Accel;
      break;

    case MT_EWOBJECTHIT:
      i = FindObject(r.ohit->ObjID);
      if(i != -1)
      { o = Objects[i];
        if(r.ohit->HitID)
          o->Damage(&Players[FindPlayer(r.ohit->HitID)]->Ship);
        RemObject(o);
        o->Destroy();
      }
      break;

    case MT_EWFWEAPONHIT:
      ammo = g_AmmoFact.Static(r.wfire->AType);
      o = ammo->ToObject();
      o->Pos.X    = r.ofhit->X,      o->Pos.Y = r.ofhit->Y;
      o->Vel.X    = r.ofhit->XV,     o->Vel.Y = r.ofhit->YV;
      o->Weight   = r.ofhit->Weight, o->Color = r.ofhit->Color;
      ammo->Extra = r.ofhit->Extra;
      o->Damage(&Players[FindPlayer(r.ofhit->HitID)]->Ship);
      break;

    case MT_EWNEWOBJECT:
      o = g_ObjFact.Create(r.nobj);
      g_World.AddObject(o);
      break;
      
    case MT_EWREMOBJECT:
      i = g_World.FindObject(r.robj->ObjID);
      if(i != -1)
      { o = g_World.Objects[i];
        g_World.RemObject(o);
        o->Destroy();
      }
      break;

    case MT_EWREMCIRCLES:
      Lock();
      Map->Lock();
      len = r.rcirc->NumCircs;
      for(i=0; i<len; i++)
        g_World.RemoveCircle(r.rcirc->Circs[i].X, r.rcirc->Circs[i].Y, r.rcirc->Circs[i].Rad);
      Map->Unlock();
      Unlock();
      break;

    case MT_EWWEAPONFIRE:
      ammo = g_AmmoFact.Create(r.wfire->AType);
      ammo->LoadFromMsg(r.wfire);
      AddObject(ammo->ToObject());
      break;

    case MT_EWWEAPONSTART:
      i = FindPlayer(r.wstart->ID);
      Players[i]->Ship.CtlFire(Players[i]->Ship.FindWep(r.wstart->WType), true);
      break;

    case MT_EWWEAPONEND:
      i = FindPlayer(r.wstart->ID);
      Players[i]->Ship.CtlFire(Players[i]->Ship.FindWep(r.wend->WType), false);
      break;

    case MT_EWWEAPONSEL:
      i = FindPlayer(r.wsel->ID);
      Players[i]->Ship.SetWeps(r.wsel->Gun, r.wsel->Special);
      break;

    case MT_EWDEATH:
      i = FindPlayer(r.death->ID);
      Players[i]->Ship.bAlive = false;
      //Players[i]->Ship.SetWeps(WT_DEFGUN, WT_DEFSPEC);
      break;

    case MT_SCSPAWN:
      Me->Ship.Pos = Me->Ship.OldPos = Vector<FLOAT>(r.spawn->X, r.spawn->Y);
      Me->Ship.Vel = Vector<FLOAT>(0, 0);
      //Me->Ship.SetWeps(WT_DEFGUN, WT_DEFSPEC);
      Me->Ship.Init();
      OldState = bNet ? WS_MULTI : WS_SINGLE;
      if(State==WS_WAITSPAWN) State=OldState;
      break;

    case MT_SCPLAYERJOIN:
      p = NEW Player;
      p->Ship.ID = r.pjoin->Player.ID;
      p->Ship.Init();
      p->Ship.Color = r.pjoin->Player.Color;
      p->Ship.SetWeps(r.pjoin->Player.Gun, r.pjoin->Player.Special);
      len = r.pjoin->Player.NameLen;
      memcpy(buf, r.pjoin->Player.Name, len);
      buf[len]=0;
      p->Name = buf;
      AddLine("%s has joined the game!", buf);
      AddPlayer(p);
      break;

    case MT_SCPLAYERLEAVE:
      p = Players[FindPlayer(r.pleave->ID)];
      AddLine("%s has left the game!", p->Name.c_str());
      RemPlayer(p);
      delete p;
      break;

    case MT_SCNETCONN:
      if(r.netconn->bFail) SetState(WS_CONNFAIL);
      else
      { place_new(buf) MsgLogin;
        s.login->VerHigh = VERMAJOR, s.login->VerLow = VERMINOR;
        s.login->NameLen = len = Name.length();
        memcpy(s.login->Name, Name.c_str(), len);
        s.login->Len += len;
        Send(s.msg, NET_SENDGUAR);
      }
      break;

    case MT_SCLOGINSUCC:
      Me = NEW Player;
      Me->Ship.ID = r.logsucc->ID;
      Me->Ship.Init();
      Me->Ship.Pos.X = r.logsucc->X, Me->Ship.Pos.Y = r.logsucc->Y;
      Me->Ship.Color = r.logsucc->Color;
      Me->Ship.SetWeps(WT_DEFGUN, WT_DEFSPEC);
      Me->Name = Name;
      nGun = WT_DEFGUN, nSpecial = WT_DEFSPEC;
      SetMe(Me);
      AddPlayer(Me);

      i  = r.logsucc->NumPlayers;
      lp = r.logsucc->Players;
      while(i--)
      { p = NEW Player;
        p->Ship.Init();
        p->Ship.ID    = lp->ID;
        p->Ship.Color = lp->Color;
        p->Ship.SetWeps(lp->Gun, lp->Special);
        len=lp->NameLen;
        memcpy(buf, lp->Name, len=lp->NameLen);
        buf[len]=0;
        p->Name = buf;
        AddPlayer(p);
        lp = (LoginPlayer*)((byte*)lp+sizeof(LoginPlayer)+len-1);
      }
      i = r.logsucc->MapLen, len = r.logsucc->BackLen;
      memcpy(c, lp, i);
      c+=i;
      *c++=0;
      memcpy(c, (char*)lp+i, len);
      c[len]=0;
      LoadLevels(buf, c);

      State = WS_MULTI;
      break;

    case MT_SCLOGINFAIL:
      Reason = r.logfail->Reason;
      State = WS_LOGINFAIL;
      Disconnect();
      break;
  }

  Unlock();
}

HRESULT WINAPI CWorld::Handler(PVOID cls, DWORD type, PVOID msg)
{ CWorld *me = (CWorld*)cls;
  if(type==DPN_MSGID_RECEIVE)
    me->MsgHandler((NetMsg*)((DPNMSG_RECEIVE *)msg)->pReceiveData);
  else if(type==DPN_MSGID_SEND_COMPLETE)
  { if(me->hMoveMsg == ((DPNMSG_SEND_COMPLETE*)msg)->hAsyncOp)
      me->hMoveMsg=NULL;
  }
  else if(type==DPN_MSGID_CONNECT_COMPLETE)
  { MsgNetConn nc;
    nc.bFail = FAILED(((DPNMSG_CONNECT_COMPLETE*)msg)->hResultCode);
    me->MsgHandler(&nc);
  }
  return S_OK;
}
